"""
This module provides utility functions that are used by the reader but generally will not be used directly by end
users.
"""

import struct
import typing

from google.protobuf import json_format
from google.protobuf.internal import containers
import lz4.block
import numpy


import redvox.api900.exceptions as exceptions
import redvox.api900.lib.api900_pb2 as api900_pb2


def calculate_uncompressed_size(buf: bytes) -> int:
    """
    Given a buffer, calculate the original size of the uncompressed packet by looking at the first four bytes.
    :param buf: Buffer where first 4 big endian bytes contain the size of the original uncompressed packet.
    :return: The total number of bytes in the original uncompressed packet.
    """
    return struct.unpack(">I", buf[:4])[0]


def uncompressed_size_bytes(size: int) -> bytes:
    """
    Given the size of the original uncompressed file, return the size as an array of 4 bytes.
    :param size: The size to convert.
    :return: The 4 bytes representing the size.
    """
    return struct.pack(">I", size)


def lz4_decompress(buf: bytes) -> bytes:
    """
    Decompresses an API 900 compressed buffer.
    :param buf: The buffer to decompress.
    :return: The uncompressed buffer.
    """
    uncompressed_size = calculate_uncompressed_size(buf)

    if uncompressed_size <= 0:
        raise exceptions.ReaderException("uncompressed size [{}] must be > 0".format(uncompressed_size))

    return lz4.block.decompress(buf[4:], uncompressed_size=calculate_uncompressed_size(buf))


# noinspection PyArgumentList
def lz4_compress(buf: bytes) -> bytes:
    """
    Compresses a buffer using LZ4. The compressed buffer is then prepended with the 4 bytes indicating the original
    size of uncompressed buffer.
    :param buf: The buffer to compress.
    :return: The compressed buffer with 4 bytes prepended indicated the original size of the uncompressed buffer.
    """
    uncompressed_size = uncompressed_size_bytes(len(buf))
    compressed = lz4.block.compress(buf, store_size=False)
    return uncompressed_size + compressed


def write_file(file: str, redvox_packet: api900_pb2.RedvoxPacket):
    """
    Writes a redvox file.  Specify the correct file type in the file string.
    :param file: str, File to write
    :param redvox_packet: protobuf packet to write
    :return: Nothing, compressed file is written to disk
    """
    buffer = lz4_compress(redvox_packet.SerializeToString())
    with open(file, "wb") as file_out:
        file_out.write(buffer)


def to_json(redvox_packet: api900_pb2.RedvoxPacket) -> str:
    """
    Converts a protobuf encoded API 900 RedVox packet into JSON.
    :param redvox_packet: The protobuf encoded packet.
    :return: A string containing JSON of this packet.
    """
    return json_format.MessageToJson(redvox_packet)


def from_json(redvox_packet_json: str) -> api900_pb2.RedvoxPacket:
    """
    Converts a JSON packet representing an API 900 packet into a protobuf encoded RedVox API 900 packet.
    :param redvox_packet_json: A string containing the json representing the packet.
    :return: A Python instance of an encoded API 900 packet.
    """
    return json_format.Parse(redvox_packet_json, api900_pb2.RedvoxPacket())


def payload_type(channel: typing.Union[api900_pb2.EvenlySampledChannel,
                                       api900_pb2.UnevenlySampledChannel]) -> str:
    """
    Given a channel, return the internal protobuf string representation of the payload's data type.
    :param channel: The channel to check the data type of.
    :return: The internal protobuf string representation of the payload's data type.
    """
    if channel is None:
        return "No channel to hold payload"

    return channel.WhichOneof("payload")

def payload_len(channel: typing.Union[api900_pb2.EvenlySampledChannel,
                                      api900_pb2.UnevenlySampledChannel]) -> int:
    """
    Given an evenly or unevenly sampled channel, extracts the entire payload.

    This will return a payload of either ints or floats and is type agnostic when it comes to the underlying
    protobuf type.
    :param channel: The protobuf channel to extract the payload from.
    :return: A numpy array of either floats or ints.
    """
    payload_type_str = payload_type(channel)

    if payload_type_str == "byte_payload":
        payload = len(channel.byte_payload.payload)
    elif payload_type_str == "uint32_payload":
        payload = len(channel.uint32_payload.payload)
    elif payload_type_str == "uint64_payload":
        payload = len(channel.uint64_payload.payload)
    elif payload_type_str == "int32_payload":
        payload = len(channel.int32_payload.payload)
    elif payload_type_str == "int64_payload":
        payload = len(channel.int64_payload.payload)
    elif payload_type_str == "float32_payload":
        payload = len(channel.float32_payload.payload)
    elif payload_type_str == "float64_payload":
        payload = len(channel.float64_payload.payload)
    else:
        payload = 0

    return payload


def extract_payload(channel: typing.Union[api900_pb2.EvenlySampledChannel,
                                          api900_pb2.UnevenlySampledChannel]) -> numpy.ndarray:
    """
    Given an evenly or unevenly sampled channel, extracts the length of the payload.

    :param channel: The protobuf channel to extract the payload from.
    :return: Length of the payload.
    """
    payload_type_str = payload_type(channel)

    if payload_type_str == "byte_payload":
        payload = channel.byte_payload.payload
        return numpy.frombuffer(payload, numpy.uint8)
    elif payload_type_str == "uint32_payload":
        payload = channel.uint32_payload.payload
    elif payload_type_str == "uint64_payload":
        payload = channel.uint64_payload.payload
    elif payload_type_str == "int32_payload":
        payload = channel.int32_payload.payload
    elif payload_type_str == "int64_payload":
        payload = channel.int64_payload.payload
    elif payload_type_str == "float32_payload":
        payload = channel.float32_payload.payload
    elif payload_type_str == "float64_payload":
        payload = channel.float64_payload.payload
    else:
        return numpy.array([])
        # raise exceptions.ReaderException("unsupported payload type {}".format(payload_type_str))

    return numpy.array(payload)


def repeated_to_list(repeated: typing.Union[containers.RepeatedCompositeFieldContainer,
                                            containers.RepeatedScalarFieldContainer]) -> typing.List:
    """
    Transforms a repeated protobuf field into a list.
    :param repeated: The repeated field to transform.
    :return: A list of the repeated items.
    """
    return repeated[0:len(repeated)]


def repeated_to_array(repeated: typing.Union[containers.RepeatedCompositeFieldContainer,
                                             containers.RepeatedScalarFieldContainer]) -> numpy.ndarray:
    """
    Transforms a repeated protobuf field into a numpy array.
    :param repeated: The repeated field to transform.
    :return: A numpy array of the repeated items.
    """
    return numpy.array(repeated_to_list(repeated))


def deinterleave_array(ndarray: numpy.ndarray, offset: int, step: int) -> numpy.ndarray:
    """
    Extracts a single channel type from an interleaved array.

    An interleaved channel contains multiple channel types in a single payload. This is useful for situations where
    a sensor produces several values for a single timestamp. For example, a GPS will produce a LATITUDE, LONGITUDE,
    ALTITUDE, and SPEED values with every update. An interleaved channel is an encoding that encodes multiple channel
    types into a single payload.

    Every channel has a field channel_types that list the channel types contained within the payload. For a GPS sensor,
    the channel_types array would look like [LATITUDE, LONGITUDE, ALTITUDE, SPEED]. The location of the channel type
    in channel_types determines the offset into the payload and the length of channel_types determines the step size.
    As such, a GPS channel payload would contain the following values:

    [LAT0, LNG0, ALT0, SPD0, LAT1, LNG1, ALT1, SPD1, ..., LATn, LNGn, ALTn, SPDn]

    This function will "deinterleave" the encoding and provide an array of a single channel type.

    :param ndarray: Interleaved array.
    :param offset: Offset into the array.
    :param step: The step size.
    :return: A numpy array of a single channel type.
    """

    if offset < 0 or offset >= len(ndarray):
        raise exceptions.ReaderException("offset {} out of range [{},{})".format(offset, 0, len(ndarray)))

    if offset >= step:
        raise exceptions.ReaderException("offset {} must be smaller than step {}".format(offset, step))

    if step <= 0 or step > len(ndarray):
        raise exceptions.ReaderException("step {} out of range [{},{})".format(step, 0, len(ndarray)))

    if len(ndarray) % step != 0:
        raise exceptions.ReaderException("step {} is not a multiple of {}".format(step, len(ndarray)))

    # pylint: disable=C1801
    if len(ndarray) == 0:
        return empty_array()

    return ndarray[offset::step]


def to_array(values: typing.Union[typing.List, numpy.ndarray]) -> numpy.ndarray:
    """
    Takes either a list or a numpy array and returns a numpy array if the parameter was a list.
    :param values: Values to convert into numpy array if values are in a list.
    :return: A numpy array of the passed in values.
    """
    if isinstance(values, list):
        return numpy.array(values)
    return values


def interleave_arrays(arrays: typing.List[numpy.ndarray]) -> numpy.ndarray:
    """
    Interleaves multiple arrays together.
    :param arrays: Arrays to interleave.
    :return: Interleaved arrays.
    """
    if len(arrays) < 2:
        raise exceptions.ReaderException("At least 2 arrays are required for interleaving")

    if len(set(map(len, arrays))) > 1:
        raise exceptions.ReaderException("all arrays must be same size")

    total_arrays = len(arrays)
    total_elements = sum(map(lambda array: array.size, arrays))
    interleaved_array = numpy.empty((total_elements,), dtype=arrays[0].dtype)
    for i in range(total_arrays):
        interleaved_array[i::total_arrays] = arrays[i]

    return interleaved_array


def implements_diff(val) -> bool:
    """
    Checks if a value implements the diff method.
    :param val: Value to check
    :return: True is the value implements diff, False otherwise.
    """
    diff_atr = getattr(val, "diff", None)
    if callable(diff_atr):
        return True

    return False


# pylint: disable=R0911
def diff(val1, val2) -> typing.Tuple[bool, typing.Optional[str]]:
    """
    Determines if the two values are different.
    :param val1: The first value to check.
    :param val2: The second value to check.
    :return: False, None if the values are the same or True, and a string displaying the differences when different.
    """
    # pylint: disable=C0123
    if type(val1) != type(val2):
        return True, "type {} != type {}".format(type(val1), type(val2))

    if implements_diff(val1) and implements_diff(val2):
        diffs = val1.diff(val2)
        if len(diffs) == 0:
            return False, None

        return True, "%s" % list(diffs)

    if isinstance(val1, numpy.ndarray) and isinstance(val2, numpy.ndarray):
        if numpy.array_equal(val1, val2):
            return False, None

        return True, "{} != {}".format(val1, val2)

    if val1 != val2:
        return True, "{} != {}".format(val1, val2)

    return False, None


def safe_index_of(lst: typing.List, val: typing.Any) -> int:
    """
    Finds the index of an item in a list and instead of throwing an exception returns -1 when the item DNE.
    :param lst: List to search through.
    :param val: The value to find the index of.
    :return: The index of the first value v found or -1.
    """
    try:
        return lst.index(val)
    except ValueError:
        return -1


def empty_array() -> numpy.ndarray:
    """Returns an empty numpy array.
    :return: An empty numpy array.
    """
    return numpy.array([])


def empty_evenly_sampled_channel() -> api900_pb2.EvenlySampledChannel:
    """
    Returns an empty protobuf EvenlySampledChannel
    :return: empty EvenlySampledChannel
    """
    obj = api900_pb2.EvenlySampledChannel()
    return obj


def empty_unevenly_sampled_channel() -> api900_pb2.UnevenlySampledChannel:
    """
    Returns an empty protobuf UnevenlySampledChannel
    :return: empty UnevenlySampledChannel
    """
    obj = api900_pb2.UnevenlySampledChannel()
    return obj


def channel_type_name_from_enum(enum_constant: int) -> str:
    """
    Returns the name of a channel type given its enumeration constant.
    :param enum_constant: The constant to turn into a name.
    :return: The name of the channel.
    """
    return api900_pb2.ChannelType.Name(enum_constant)


def get_metadata(metadata: typing.List[str], k: str) -> str:
    """
    Given a meta-data key, extract the value.
    :param metadata: List of metadata to extract value from.
    :param k: The meta-data key.
    :return: The value corresponding to the key or an empty string.
    """
    if len(metadata) % 2 != 0:
        raise exceptions.ReaderException("metadata list must contain an even number of items")

    idx = safe_index_of(metadata, k)
    if idx < 0:
        return ""

    return metadata[idx + 1]


def get_metadata_as_dict(metadata: typing.List[str]) -> typing.Dict[str, str]:
    """
    Since the metadata is inherently key-value, it may be useful to turn the metadata list into a python dictionary.
    :param metadata: The metadata list.
    :return: Metadata as a python dictionary.
    """
    if not metadata:
        return {}

    if len(metadata) % 2 != 0:
        raise exceptions.ReaderException("metadata list must contain an even number of items")

    metadata_dict = {}
    metadata_copy = metadata.copy()
    while len(metadata_copy) >= 2:
        metadata_key = metadata_copy.pop(0)
        metadata_value = metadata_copy.pop(0)
        if metadata_key not in metadata_dict:
            metadata_dict[metadata_key] = metadata_value
    return metadata_dict


def metadata_dict_to_list(metadata_dict: typing.Dict[str, str]) -> typing.List[str]:
    """
    Converts a dictionary containing metadata into a list of metadata.
    :param metadata_dict: The dictionary of metadata.
    :return: A list of metadata.
    """
    metadata_list = []
    for key, value in metadata_dict.items():
        metadata_list.extend([key, value])
    return metadata_list



"""
Provides global settings that tweak the inner workings of the SDK.
"""

import os
from typing import Optional

REDVOX_ENABLE_PARALLELISM_ENV: str = "REDVOX_ENABLE_PARALLELISM"


def is_parallelism_enabled_env() -> Optional[bool]:
    """
    Tests if parallelism is enabled or disabled by checking the presence and value of an environmental variable.
    If the env var DNE, None is returned. If the env var does exist, it is parsed as either "true" or "false" into
    the corresponding boolean and returned. If the env var exists, but is not one of "true" or "false", None is
    returned.
    :return: Either True or False if the env var exists and can be parsed or None.
    """
    if REDVOX_ENABLE_PARALLELISM_ENV not in os.environ:
        return None

    env_val: str = os.environ.get(REDVOX_ENABLE_PARALLELISM_ENV).lower()
    if env_val == "true":
        return True
    elif env_val == "false":
        return False
    else:
        return None


__PARALLELISM_ENABLED: Optional[bool] = is_parallelism_enabled_env()


def set_parallelism_enabled(parallelism_enabled: bool) -> None:
    """
    Sets whether parallelism is enabled or disabled within the SDK.
    :param parallelism_enabled: True to enable, False otherwise
    """
    global __PARALLELISM_ENABLED
    __PARALLELISM_ENABLED = parallelism_enabled


def is_parallelism_enabled() -> bool:
    """
    Returns whether or not parallelism is enabled with the SDK.
    :return: Whether or not parallelism is enabled with the SDK.
    """
    global __PARALLELISM_ENABLED
    if __PARALLELISM_ENABLED is None:
        __PARALLELISM_ENABLED = is_parallelism_enabled_env()
    return False if __PARALLELISM_ENABLED is None else __PARALLELISM_ENABLED



"""
This modules contains constants used throughout the SDK.
"""

import enum


class PayloadType(enum.Enum):
    """
    This class provides type safe constants for API 900 payload types.
    """
    BYTE_PAYLOAD: str = "byte_payload"
    UINT32_PAYLOAD: str = "uint32_payload"
    UINT64_PAYLOAD: str = "uint64_payload"
    INT32_PAYLOAD: str = "int32_payload"
    INT64_PAYLOAD: str = "int64_payload"
    FLOAT32_PAYLOAD: str = "float32_payload"
    FLOAT64_PAYLOAD: str = "float64_payload"



from setuptools import setup, find_packages

import redvox

with open("requirements.txt", "r") as requirements_file:
    requirements = list(map(lambda line: line.strip(), requirements_file.readlines()))
    requirements = list(filter(lambda line: (not line.startswith("#")) and len(line) > 0, requirements))

setup(name=redvox.NAME,
      version=redvox.VERSION,
      url='https://bitbucket.org/redvoxhi/redvox-api900-python-reader/src/master/',
      license='Apache',
      author='RedVox',
      author_email='dev@redvoxsound.com',
      description='Library for working with RedVox files.',
      packages=find_packages(include=["redvox",
                                      "redvox.api900",
                                      "redvox.api900.lib",
                                      "redvox.api900.sensors",
                                      "redvox.api900.timesync",
                                      "redvox.api900.qa",
                                      "redvox.api1000",
                                      "redvox.api1000.common",
                                      "redvox.api1000.gui",
                                      "redvox.api1000.gui.data_explorer",
                                      "redvox.api1000.proto",
                                      "redvox.api1000.wrapped_redvox_packet",
                                      "redvox.api1000.wrapped_redvox_packet.sensors",
                                      "redvox.api1000.wrapped_redvox_packet.sensors.derived",
                                      "redvox.cli",
                                      "redvox.cloud",
                                      "redvox.common",
                                      "redvox.common.gui",
                                      ],
                             exclude=['tests']),
      long_description=open('README.md').read(),
      install_requires=requirements,
      extras_require={
          "GUI":  ["PySide6>=6.0.2", "matplotlib>=3.2.2"],
      },
      entry_points={
            'console_scripts': [
                'redvox-cli=redvox.cli.cli:main',
            ]
      },
      python_requires=">=3.6")



"""
This module contains functionality for working with XYZ channeled API M sensors.
"""

from typing import List, Optional

import redvox.api1000.common.common as common
import redvox.api1000.common.typing
import redvox.api1000.proto.redvox_api_m_pb2 as redvox_api_m_pb2
import redvox.api1000.common.generic


class Xyz(
    redvox.api1000.common.generic.ProtoBase[redvox_api_m_pb2.RedvoxPacketM.Sensors.Xyz]
):
    """
    Encapsulates a sensor with x, y, and z data and metadata
    """

    def __init__(self, proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.Xyz):
        super().__init__(proto)
        self._timestamps: common.TimingPayload = common.TimingPayload(proto.timestamps)
        self._x_samples: common.SamplePayload = common.SamplePayload(proto.x_samples)
        self._y_samples: common.SamplePayload = common.SamplePayload(proto.y_samples)
        self._z_samples: common.SamplePayload = common.SamplePayload(proto.z_samples)

    @staticmethod
    def new() -> "Xyz":
        """
        :return: A new, empty Xyz sensor instance
        """
        return Xyz(redvox_api_m_pb2.RedvoxPacketM.Sensors.Xyz())

    def set_unit_xyz(self, unit: common.Unit) -> "Xyz":
        """
        Sets the unit of the x, y, and z channels
        :param unit: Unit to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(unit, [common.Unit])
        self._x_samples.set_unit(unit)
        self._y_samples.set_unit(unit)
        self._z_samples.set_unit(unit)
        return self

    def get_sensor_description(self) -> str:
        """
        :return: This sensor's description
        """
        return self._proto.sensor_description

    def set_sensor_description(self, sensor_description: str) -> "Xyz":
        """
        Sets this sensor's description
        :param sensor_description: Description to set
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sensor_description, [str])
        self._proto.sensor_description = sensor_description
        return self

    def get_timestamps(self) -> common.TimingPayload:
        """
        :return: TimingPayload with timestamps associated with each sample
        """
        return self._timestamps

    def set_timestamps(self, timestamps: common.TimingPayload) -> "Xyz":
        """
        Sets the timestamps.
        :param timestamps: Timestamps to set.
        :return: A modified instance of self.
        """
        common.check_type(timestamps, [common.TimingPayload])
        self.get_proto().timestamps.CopyFrom(timestamps.get_proto())
        self._timestamps = common.TimingPayload(self.get_proto().timestamps)
        return self

    def get_x_samples(self) -> common.SamplePayload:
        """
        :return: SamplePayload associated with x channel
        """
        return self._x_samples

    def set_x_samples(self, x_samples: common.SamplePayload) -> "Xyz":
        """
        Sets the X channel samples.
        :param x_samples: Samples to set.
        :return: A modified instance of self.
        """
        common.check_type(x_samples, [common.SamplePayload])
        # noinspection Mypy
        self.get_proto().x_samples.CopyFrom(x_samples.get_proto())
        self._x_samples = common.SamplePayload(self.get_proto().x_samples)
        return self

    def get_y_samples(self) -> common.SamplePayload:
        """
        :return: SamplePayload associated with y channel
        """
        return self._y_samples

    def set_y_samples(self, y_samples: common.SamplePayload) -> "Xyz":
        """
        Sets the Y channel samples.
        :param y_samples: Samples to set.
        :return: A modified instance of self.
        """
        common.check_type(y_samples, [common.SamplePayload])
        # noinspection Mypy
        self.get_proto().x_samples.CopyFrom(y_samples.get_proto())
        self._y_samples = common.SamplePayload(self.get_proto().x_samples)
        return self

    def get_z_samples(self) -> common.SamplePayload:
        """
        :return: SamplePayload associated with z channel
        """
        return self._z_samples

    def set_z_samples(self, z_samples: common.SamplePayload) -> "Xyz":
        """
        Sets the Z channel samples.
        :param z_samples: Samples to set.
        :return: A modified instance of self.
        """
        common.check_type(z_samples, [common.SamplePayload])
        # noinspection Mypy
        self.get_proto().x_samples.CopyFrom(z_samples.get_proto())
        self._z_samples = common.SamplePayload(self.get_proto().x_samples)
        return self


def validate_xyz(
    xyz_sensor: Xyz, payload_unit: Optional[common.Unit] = None
) -> List[str]:
    """
    Validates the XYZ sensor.
    :param xyz_sensor: Sensor to validate.
    :param payload_unit: A list of validation errors.
    :return:
    """
    errors_list = common.validate_timing_payload(xyz_sensor.get_timestamps())
    errors_list.extend(
        common.validate_sample_payload(
            xyz_sensor.get_x_samples(),
            xyz_sensor.get_sensor_description(),
            payload_unit,
        )
    )
    errors_list.extend(
        common.validate_sample_payload(
            xyz_sensor.get_y_samples(),
            xyz_sensor.get_sensor_description(),
            payload_unit,
        )
    )
    errors_list.extend(
        common.validate_sample_payload(
            xyz_sensor.get_z_samples(),
            xyz_sensor.get_sensor_description(),
            payload_unit,
        )
    )
    return errors_list



"""
This module loads sensor data from Redvox packets
"""

from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from redvox.common.stats_helper import StatsContainer
from redvox.common import date_time_utils as dtu
from redvox.common import gap_and_pad_utils as gpu
from redvox.common.sensor_data import SensorType, SensorData
from redvox.api1000.wrapped_redvox_packet.sensors import xyz, single, audio, image, location
from redvox.api1000.wrapped_redvox_packet.station_information import NetworkType, CellServiceState, PowerState
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM


# Dataframe column definitions
COMPRESSED_AUDIO_COLUMNS = ["timestamps", "unaltered_timestamps", "compressed_audio", "audio_codec"]
IMAGE_COLUMNS = ["timestamps", "unaltered_timestamps", "image", "image_codec"]
LOCATION_COLUMNS = ["timestamps", "unaltered_timestamps", "gps_timestamps", "latitude", "longitude",
                    "altitude", "speed", "bearing", "horizontal_accuracy", "vertical_accuracy",
                    "speed_accuracy", "bearing_accuracy", "location_provider"]


def get_empty_sensor_data(
        name: str, sensor_type: SensorType = SensorType.UNKNOWN_SENSOR
) -> SensorData:
    """
    create a sensor data object with no data
    :param name: name of the sensor
    :param sensor_type: type of the sensor to create, default SensorType.UNKNOWN_SENSOR
    :return: empty sensor
    """
    return SensorData(name, pd.DataFrame([], columns=["timestamps"]), sensor_type)


def get_sensor_description(sensor: Union[xyz.Xyz, single.Single, audio.Audio, image.Image, location.Location]) -> str:
    """
    read the sensor's description from the sensor
    :param sensor: the sensor to read the description from
    :return: the sensor's description
    """
    return sensor.get_sensor_description()


def get_sensor_description_list(wrapped_packets: List[WrappedRedvoxPacketM], sensor_type: SensorType) -> str:
    """
    read the sensor_type sensor's description from a list of packets
    :param wrapped_packets: the list of packets to read from
    :param sensor_type: the SensorType of the sensor to read the description of
    :return: the sensor_type sensor's description
    """
    for packet in wrapped_packets:
        if sensor_type == SensorType.AUDIO and packet.get_sensors().has_audio():
            return get_sensor_description(packet.get_sensors().get_audio())
        if sensor_type == SensorType.IMAGE and packet.get_sensors().has_image():
            return get_sensor_description(packet.get_sensors().get_image())
        if sensor_type == SensorType.LOCATION and packet.get_sensors().has_location():
            return get_sensor_description(packet.get_sensors().get_location())
        if sensor_type == SensorType.PRESSURE and packet.get_sensors().has_pressure():
            return get_sensor_description(packet.get_sensors().get_pressure())
        if sensor_type == SensorType.ACCELEROMETER and packet.get_sensors().has_accelerometer():
            return get_sensor_description(packet.get_sensors().get_accelerometer())
        if sensor_type == SensorType.AMBIENT_TEMPERATURE and packet.get_sensors().has_ambient_temperature():
            return get_sensor_description(packet.get_sensors().get_ambient_temperature())
        if sensor_type == SensorType.COMPRESSED_AUDIO and packet.get_sensors().has_compressed_audio():
            return get_sensor_description(packet.get_sensors().get_compressed_audio())
        if sensor_type == SensorType.GRAVITY and packet.get_sensors().has_gravity():
            return get_sensor_description(packet.get_sensors().get_gravity())
        if sensor_type == SensorType.GYROSCOPE and packet.get_sensors().has_gyroscope():
            return get_sensor_description(packet.get_sensors().get_gyroscope())
        if sensor_type == SensorType.LIGHT and packet.get_sensors().has_light():
            return get_sensor_description(packet.get_sensors().get_light())
        if sensor_type == SensorType.LINEAR_ACCELERATION and packet.get_sensors().has_linear_acceleration():
            return get_sensor_description(packet.get_sensors().get_linear_acceleration())
        if sensor_type == SensorType.MAGNETOMETER and packet.get_sensors().has_magnetometer():
            return get_sensor_description(packet.get_sensors().get_magnetometer())
        if sensor_type == SensorType.ORIENTATION and packet.get_sensors().has_orientation():
            return get_sensor_description(packet.get_sensors().get_orientation())
        if sensor_type == SensorType.PROXIMITY and packet.get_sensors().has_proximity():
            return get_sensor_description(packet.get_sensors().get_proximity())
        if sensor_type == SensorType.RELATIVE_HUMIDITY and packet.get_sensors().has_relative_humidity():
            return get_sensor_description(packet.get_sensors().get_relative_humidity())
        if sensor_type == SensorType.ROTATION_VECTOR and packet.get_sensors().has_rotation_vector():
            return get_sensor_description(packet.get_sensors().get_rotation_vector())


def get_sample_statistics(data_df: pd.DataFrame) -> Tuple[float, float, float]:
    """
    calculate the sample rate, interval and interval std dev using the timestamps in the dataframe
    :param data_df: the dataframe containing timestamps to calculate statistics from
    :return: a Tuple containing the sample rate, interval and interval std dev
    """
    if data_df["timestamps"].size > 1:
        sample_interval = dtu.microseconds_to_seconds(
            float(np.mean(np.diff(data_df["timestamps"])))
        )
        sample_interval_std = dtu.microseconds_to_seconds(
            float(np.std(np.diff(data_df["timestamps"])))
        )
    else:
        sample_interval = np.nan
        sample_interval_std = np.nan
    return 1 / sample_interval, sample_interval, sample_interval_std


def read_apim_xyz_sensor(sensor: xyz.Xyz, column_id: str) -> pd.DataFrame:
    """
    read a sensor that has xyz data channels from an api M data packet
    raises Attribute Error if sensor does not contain xyz channels
    :param sensor: the xyz api M sensor to read
    :param column_id: string, used to name the columns
    :return: Dataframe representing the data in the sensor
    """
    timestamps = sensor.get_timestamps().get_timestamps()
    try:
        columns = ["timestamps", "unaltered_timestamps", f"{column_id}_x", f"{column_id}_y", f"{column_id}_z"]
        return pd.DataFrame(
            np.transpose(
                [
                    timestamps,
                    timestamps,
                    sensor.get_x_samples().get_values(),
                    sensor.get_y_samples().get_values(),
                    sensor.get_z_samples().get_values(),
                ]
            ),
            columns=columns,
        )
    except AttributeError:
        raise


def load_apim_xyz_sensor(sensor_type: SensorType, data: List[List[float]],
                         gaps: List[Tuple[float, float]], column_name: str,
                         description: str, sample_interval_micros: float) -> Optional[SensorData]:
    """
    create a three channel sensor of sensor_type with the column name, description, and data given.
    :param sensor_type: the SensorType of sensor to create
    :param data: the list of data to be added; requires timestamps to be the first list
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :param column_name: the name of the columns that contain the data
    :param description: the description of the sensor
    :param sample_interval_micros: known sample interval in microseconds
    :return: SensorData object or None if no data
    """
    if len(data[0]) > 0:
        data_df = pd.DataFrame(np.transpose([data[0], data[0], data[1], data[2], data[3]]),
                               columns=["timestamps", "unaltered_timestamps",
                                        f"{column_name}_x", f"{column_name}_y", f"{column_name}_z"])
        return SensorData(
            description,
            gpu.fill_gaps(data_df, gaps, sample_interval_micros),
            sensor_type,
            calculate_stats=True
        )
    return None


def read_apim_single_sensor(sensor: single.Single, column_id: str) -> pd.DataFrame:
    """
    read a sensor that has a single data channel from an api M data packet
    raises Attribute Error if sensor does not contain exactly one data channel
    :param sensor: the single channel api M sensor to read
    :param column_id: string, used to name the columns
    :return: Dataframe representing the data in the sensor
    """
    timestamps = sensor.get_timestamps().get_timestamps()
    try:
        columns = ["timestamps", "unaltered_timestamps", column_id]
        return pd.DataFrame(
            np.transpose([timestamps, timestamps, sensor.get_samples().get_values()]),
            columns=columns,
        )
    except AttributeError:
        raise


def load_apim_single_sensor(sensor_type: SensorType, timestamps: List[float], data: List[float],
                            gaps: List[Tuple[float, float]], column_name: str,
                            description: str, sample_interval_micros: float) -> Optional[SensorData]:
    """
    Create a single channel sensor of sensor_type with the column name, timestamps, data, and description
    :param sensor_type: the SensorType of sensor to create
    :param timestamps: the timestamps of the data
    :param data: the list of data to be added
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :param column_name: the name of the columns that contain the data
    :param description: the description of the sensor
    :param sample_interval_micros: known sample interval in microseconds
    :return:
    """
    if len(timestamps) > 0:
        data_df = pd.DataFrame(np.transpose([timestamps, timestamps, data]),
                               columns=["timestamps", "unaltered_timestamps", column_name])
        return SensorData(
            description,
            gpu.fill_gaps(data_df, gaps, sample_interval_micros),
            sensor_type,
            calculate_stats=True
        )
    return None


def load_apim_audio(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load audio data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: audio sensor data if it exists, None otherwise
    """
    audio_sensor = wrapped_packet.get_sensors().get_audio()
    if audio_sensor and wrapped_packet.get_sensors().validate_audio():
        sample_rate_hz = audio_sensor.get_sample_rate()
        data_for_df = audio_sensor.get_samples().get_values()
        timestamps = gpu.calc_evenly_sampled_timestamps(
            audio_sensor.get_first_sample_timestamp(), audio_sensor.get_num_samples(),
            dtu.seconds_to_microseconds(1/sample_rate_hz)
        )
        return SensorData(
            get_sensor_description(audio_sensor),
            pd.DataFrame(np.transpose([timestamps, timestamps, data_for_df]), columns=gpu.AUDIO_DF_COLUMNS),
            SensorType.AUDIO,
            sample_rate_hz,
            1 / sample_rate_hz,
            0.0,
            True,
            )
    return None


def load_apim_audio_from_list(wrapped_packets: List[WrappedRedvoxPacketM]) \
        -> (Optional[SensorData], List[Tuple[float, float]]):
    """
    load audio data from a list of wrapped packets
    NOTE: This only works because audio sensors in the list should all have the same number of data points.
    :param wrapped_packets: packets with data to load
    :return: audio sensor data if it exists, None otherwise
    """
    if len(wrapped_packets) > 0:
        if wrapped_packets[0].get_sensors().get_audio() and wrapped_packets[0].get_sensors().validate_audio():
            try:
                sample_rate_hz = wrapped_packets[0].get_sensors().get_audio().get_sample_rate()
                packet_info = [(p.get_sensors().get_audio().get_first_sample_timestamp(),
                                p.get_sensors().get_audio().get_samples().get_values(),
                                p.get_sensors().get_audio().get_num_samples())
                               for p in wrapped_packets]
                df, gaps = gpu.fill_audio_gaps(packet_info, dtu.seconds_to_microseconds(1/sample_rate_hz))

                return SensorData(
                    get_sensor_description_list(wrapped_packets, SensorType.AUDIO),
                    df,
                    SensorType.AUDIO,
                    sample_rate_hz,
                    1 / sample_rate_hz,
                    0.0,
                    True,
                    ), gaps
            except ValueError as error:
                print("Error occurred while loading audio data for station "
                      f"{wrapped_packets[0].get_station_information().get_id()}.\n"
                      f"Original error message: {error}")
    return None, []


def load_apim_compressed_audio(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load compressed audio data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: compressed audio sensor data if it exists, None otherwise
    """
    comp_audio = wrapped_packet.get_sensors().get_compressed_audio()
    if comp_audio and wrapped_packet.get_sensors().validate_compressed_audio():
        sample_rate_hz = comp_audio.get_sample_rate()
        return SensorData(
            get_sensor_description(comp_audio),
            pd.DataFrame(
                np.transpose(
                    [
                        comp_audio.get_first_sample_timestamp(),
                        comp_audio.get_first_sample_timestamp(),
                        comp_audio.get_audio_bytes(),
                        comp_audio.get_audio_codec(),
                    ]
                ),
                columns=COMPRESSED_AUDIO_COLUMNS,
            ),
            SensorType.COMPRESSED_AUDIO,
            sample_rate_hz,
            1 / sample_rate_hz,
            0.0,
            True,
            )
    return None


def load_apim_compressed_audio_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                         gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load compressed audio data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: compressed audio sensor data if it exists, None otherwise
    """
    data_list = [[], [], []]
    for packet in wrapped_packets:
        comp_audio = packet.get_sensors().get_compressed_audio()
        if comp_audio and packet.get_sensors().validate_compressed_audio():
            data_list[0].append(comp_audio.get_first_sample_timestamp())
            data_list[1].append(comp_audio.get_audio_bytes())
            data_list[2].append(comp_audio.get_audio_codec())
    if len(data_list[0]) > 0:
        data_df = gpu.fill_gaps(pd.DataFrame(
            np.tranpose([data_list[0], data_list[0], data_list[1], data_list[2]]),
            columns=COMPRESSED_AUDIO_COLUMNS), gaps,
            dtu.seconds_to_microseconds(wrapped_packets[0].get_packet_duration_s()))
        data_df["audio_codec"] = [audio.AudioCodec(d) for d in data_df["audio_codec"]]
        sample_rate_hz = wrapped_packets[0].get_sensors().get_compressed_audio().get_sample_rate()
        return SensorData(
            get_sensor_description_list(wrapped_packets, SensorType.COMPRESSED_AUDIO),
            data_df,
            SensorType.COMPRESSED_AUDIO,
            sample_rate_hz,
            1 / sample_rate_hz,
            0.0,
            True,
            )
    return None


def load_apim_image(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load image data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: image sensor data if it exists, None otherwise
    """
    image_sensor = wrapped_packet.get_sensors().get_image()
    if image_sensor and wrapped_packet.get_sensors().validate_image():
        timestamps = image_sensor .get_timestamps().get_timestamps()
        codecs = np.full(len(timestamps), image_sensor .get_image_codec().value)
        data_df = pd.DataFrame(
            np.transpose([timestamps, timestamps, image_sensor .get_samples(), codecs]),
            columns=IMAGE_COLUMNS,
        )
        data_df["image_codec"] = [image.ImageCodec(d) for d in data_df["image_codec"]]
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(image_sensor),
            data_df,
            SensorType.IMAGE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )
    return None


def load_apim_image_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                              gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load image data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: image sensor data if it exists, None otherwise
    """
    data_list = [[], [], []]
    for packet in wrapped_packets:
        image_sensor = packet.get_sensors().get_image()
        if image_sensor and packet.get_sensors().validate_image():
            data_list[0].extend(image_sensor.get_timestamps().get_timestamps())
            data_list[1].extend(image_sensor.get_samples())
            data_list[2].extend([image_sensor.get_image_codec() for i in range(image_sensor.get_num_images())])
    if len(data_list[0]) > 0:
        # image is collected 1 per packet or 1 per second
        if len(data_list[0]) > len(wrapped_packets):
            sample_rate = 1.0
        else:
            sample_rate = 1 / wrapped_packets[0].get_packet_duration_s()
        sample_interval = 1 / sample_rate
        sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(data_list[0]))))\
            if len(data_list[0]) > 1 else np.nan
        return SensorData(
            get_sensor_description_list(wrapped_packets, SensorType.IMAGE),
            gpu.fill_gaps(pd.DataFrame(
                np.transpose([data_list[0], data_list[0], data_list[1], data_list[2]]),
                columns=IMAGE_COLUMNS), gaps, dtu.seconds_to_microseconds(sample_interval)),
            SensorType.IMAGE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )
    return None


def load_apim_location(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load location data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: location sensor data if it exists, None otherwise
    """
    loc = wrapped_packet.get_sensors().get_location()
    if loc and wrapped_packet.get_sensors().validate_location():
        if loc.is_only_best_values():
            if loc.get_last_best_location():
                best_loc = loc.get_last_best_location()
            else:
                best_loc = loc.get_overall_best_location()
            data_for_df = [
                [
                    best_loc.get_latitude_longitude_timestamp().get_mach(),
                    best_loc.get_latitude_longitude_timestamp().get_mach(),
                    best_loc.get_latitude_longitude_timestamp().get_gps(),
                    best_loc.get_latitude(),
                    best_loc.get_longitude(),
                    best_loc.get_altitude(),
                    best_loc.get_speed(),
                    best_loc.get_bearing(),
                    best_loc.get_horizontal_accuracy(),
                    best_loc.get_vertical_accuracy(),
                    best_loc.get_speed_accuracy(),
                    best_loc.get_bearing_accuracy(),
                    best_loc.get_location_provider(),
                ]
            ]
        else:
            timestamps = loc.get_timestamps().get_timestamps()
            if len(timestamps) > 0:
                gps_timestamps = loc.get_timestamps_gps().get_timestamps()
                lat_samples = loc.get_latitude_samples().get_values()
                lon_samples = loc.get_longitude_samples().get_values()
                alt_samples = loc.get_altitude_samples().get_values()
                spd_samples = loc.get_speed_samples().get_values()
                bear_samples = loc.get_bearing_samples().get_values()
                hor_acc_samples = loc.get_horizontal_accuracy_samples().get_values()
                vert_acc_samples = loc.get_vertical_accuracy_samples().get_values()
                spd_acc_samples = loc.get_speed_accuracy_samples().get_values()
                bear_acc_samples = loc.get_bearing_accuracy_samples().get_values()
                loc_prov_samples = loc.get_location_providers().get_values()
                data_for_df = []
                for i in range(len(timestamps)):
                    new_entry = [
                        timestamps[i],
                        timestamps[i],
                        np.nan if len(gps_timestamps) <= i else gps_timestamps[i],
                        lat_samples[i],
                        lon_samples[i],
                        np.nan if len(alt_samples) <= i else alt_samples[i],
                        np.nan if len(spd_samples) <= i else spd_samples[i],
                        np.nan if len(bear_samples) <= i else bear_samples[i],
                        np.nan if len(hor_acc_samples) <= i else hor_acc_samples[i],
                        np.nan if len(vert_acc_samples) <= i else vert_acc_samples[i],
                        np.nan if len(spd_acc_samples) <= i else spd_acc_samples[i],
                        np.nan if len(bear_acc_samples) <= i else bear_acc_samples[i],
                        np.nan if len(loc_prov_samples) <= i else loc_prov_samples[i],
                    ]
                    data_for_df.append(new_entry)
            else:
                return None
        data_df = pd.DataFrame(data_for_df, columns=LOCATION_COLUMNS)
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(loc),
            data_df,
            SensorType.LOCATION,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )
    return None


def load_apim_location_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                 gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load location data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: location sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], [], [], [], [], [], [], [], [], []]
    loc_stats = StatsContainer("location_sensor")
    for packet in wrapped_packets:
        loc = packet.get_sensors().get_location()
        if loc and packet.get_sensors().validate_location():
            if loc.is_only_best_values():
                if loc.get_last_best_location():
                    best_loc = loc.get_last_best_location()
                else:
                    best_loc = loc.get_overall_best_location()
                data_list[0].append(best_loc.get_latitude_longitude_timestamp().get_mach())
                data_list[1].append(best_loc.get_latitude_longitude_timestamp().get_gps())
                data_list[2].append(best_loc.get_latitude())
                data_list[3].append(best_loc.get_longitude())
                data_list[4].append(best_loc.get_altitude())
                data_list[5].append(best_loc.get_speed())
                data_list[6].append(best_loc.get_bearing())
                data_list[7].append(best_loc.get_horizontal_accuracy())
                data_list[8].append(best_loc.get_vertical_accuracy())
                data_list[9].append(best_loc.get_speed_accuracy())
                data_list[10].append(best_loc.get_bearing_accuracy())
                data_list[11].append(best_loc.get_location_provider())
                loc_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                num_samples = loc.get_timestamps().get_timestamps_count()
                if num_samples > 0:
                    samples = loc.get_timestamps().get_timestamps()
                    data_list[0].extend(samples)
                    if num_samples == 1:
                        loc_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
                    else:
                        loc_stats.add(np.mean(np.diff(samples)), np.std(np.diff(samples)), num_samples - 1)
                    # data_list[4].extend([np.nan if len(samples) < i + 1
                    #                      else samples[i] for i in range(num_samples)])
                    for i in range(num_samples):
                        samples = loc.get_timestamps_gps().get_timestamps()
                        data_list[1].append(np.nan if len(samples) <= i else samples[i])
                        samples = loc.get_latitude_samples().get_values()
                        data_list[2].append(samples[i])
                        samples = loc.get_longitude_samples().get_values()
                        data_list[3].append(samples[i])
                        samples = loc.get_altitude_samples().get_values()
                        data_list[4].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_speed_samples().get_values()
                        data_list[5].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_bearing_samples().get_values()
                        data_list[6].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_horizontal_accuracy_samples().get_values()
                        data_list[7].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_vertical_accuracy_samples().get_values()
                        data_list[8].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_speed_accuracy_samples().get_values()
                        data_list[9].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_bearing_accuracy_samples().get_values()
                        data_list[10].append(np.nan if len(samples) < i + 1 else samples[i])
                        samples = loc.get_location_providers().get_values()
                        data_list[11].append(location.LocationProvider["UNKNOWN"]
                                             if len(samples) < i + 1 else samples[i])
    if len(data_list[0]) > 0:
        data_list.insert(1, data_list[0].copy())
        return SensorData(
            get_sensor_description_list(wrapped_packets, SensorType.LOCATION),
            gpu.fill_gaps(pd.DataFrame(np.transpose(data_list), columns=LOCATION_COLUMNS), gaps,
                          loc_stats.mean_of_means()),
            SensorType.LOCATION,
            calculate_stats=True
        )
    return None


def load_apim_pressure(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load pressure data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: pressure sensor data if it exists, None otherwise
    """
    pressure = wrapped_packet.get_sensors().get_pressure()
    if pressure and wrapped_packet.get_sensors().validate_pressure():
        data_df = read_apim_single_sensor(pressure, "pressure")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(pressure),
            data_df,
            SensorType.PRESSURE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_pressure_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                 gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load pressure data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: pressure sensor data if it exists, None otherwise
    """
    data_list = []
    timestamps = []
    pressure_stats = StatsContainer("pressure_sensor")
    for packet in wrapped_packets:
        pressure = packet.get_sensors().get_pressure()
        if pressure and packet.get_sensors().validate_pressure():
            data_list.extend(pressure.get_samples().get_values())
            ts = pressure.get_timestamps().get_timestamps()
            timestamps.extend(ts)
            if pressure.get_timestamps().get_timestamps_count() == 1:
                pressure_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                pressure_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list) > 0:
        return load_apim_single_sensor(SensorType.PRESSURE, timestamps, data_list, gaps, "pressure",
                                       get_sensor_description_list(wrapped_packets, SensorType.PRESSURE),
                                       pressure_stats.mean_of_means())
    return None


def load_apim_light(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load light data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: light sensor data if it exists, None otherwise
    """
    light = wrapped_packet.get_sensors().get_light()
    if light and wrapped_packet.get_sensors().validate_light():
        data_df = read_apim_single_sensor(light, "light")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(light),
            data_df,
            SensorType.LIGHT,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_light_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                              gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load light data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: light sensor data if it exists, None otherwise
    """
    data_list = []
    timestamps = []
    light_stats = StatsContainer("light_sensor")
    for packet in wrapped_packets:
        light = packet.get_sensors().get_light()
        if light and packet.get_sensors().validate_light():
            data_list.extend(light.get_samples().get_values())
            ts = light.get_timestamps().get_timestamps()
            timestamps.extend(ts)
            if light.get_timestamps().get_timestamps_count() == 1:
                light_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                light_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list) > 0:
        return load_apim_single_sensor(SensorType.LIGHT, timestamps, data_list, gaps, "light",
                                       get_sensor_description_list(wrapped_packets, SensorType.LIGHT),
                                       light_stats.mean_of_means())
    return None


def load_apim_proximity(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load proximity data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: proximity sensor data if it exists, None otherwise
    """
    proximity = wrapped_packet.get_sensors().get_proximity()
    if proximity and wrapped_packet.get_sensors().validate_proximity():
        data_df = read_apim_single_sensor(proximity, "proximity")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(proximity),
            data_df,
            SensorType.PROXIMITY,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_proximity_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                  gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load proximity data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: proximity sensor data if it exists, None otherwise
    """
    data_list = []
    timestamps = []
    proximity_stats = StatsContainer("proximity_sensor")
    for packet in wrapped_packets:
        proximity = packet.get_sensors().get_proximity()
        if proximity and packet.get_sensors().validate_proximity():
            data_list.extend(proximity.get_samples().get_values())
            ts = proximity.get_timestamps().get_timestamps()
            timestamps.extend(ts)
            if proximity.get_timestamps().get_timestamps_count() == 1:
                proximity_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                proximity_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list) > 0:
        return load_apim_single_sensor(SensorType.PROXIMITY, timestamps, data_list, gaps, "proximity",
                                       get_sensor_description_list(wrapped_packets, SensorType.PROXIMITY),
                                       proximity_stats.mean_of_means())
    return None


def load_apim_ambient_temp(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load ambient temperature data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: ambient temperature sensor data if it exists, None otherwise
    """
    ambient_temp = wrapped_packet.get_sensors().get_ambient_temperature()
    if ambient_temp and wrapped_packet.get_sensors().validate_ambient_temperature():
        data_df = read_apim_single_sensor(ambient_temp, "ambient_temp")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(ambient_temp),
            data_df,
            SensorType.AMBIENT_TEMPERATURE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_ambient_temp_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                     gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load ambient temperature data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: ambient temperature sensor data if it exists, None otherwise
    """
    data_list = []
    timestamps = []
    amb_temp_stats = StatsContainer("amb_temp_sensor")
    for packet in wrapped_packets:
        amb_temp = packet.get_sensors().get_ambient_temperature()
        if amb_temp and packet.get_sensors().validate_ambient_temperature():
            data_list.extend(amb_temp.get_samples().get_values())
            ts = amb_temp.get_timestamps().get_timestamps()
            timestamps.extend(ts)
            if amb_temp.get_timestamps().get_timestamps_count() == 1:
                amb_temp_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                amb_temp_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list) > 0:
        return load_apim_single_sensor(SensorType.AMBIENT_TEMPERATURE, timestamps, data_list, gaps, "ambient_temp",
                                       get_sensor_description_list(wrapped_packets, SensorType.AMBIENT_TEMPERATURE),
                                       amb_temp_stats.mean_of_means())
    return None


def load_apim_rel_humidity(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load relative humidity data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: relative humidity sensor data if it exists, None otherwise
    """
    rel_humidity = wrapped_packet.get_sensors().get_relative_humidity()
    if rel_humidity and wrapped_packet.get_sensors().validate_relative_humidity():
        data_df = read_apim_single_sensor(rel_humidity, "rel_humidity")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(rel_humidity),
            data_df,
            SensorType.RELATIVE_HUMIDITY,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_rel_humidity_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                     gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load relative humidity data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: relative humidity sensor data if it exists, None otherwise
    """
    data_list = []
    timestamps = []
    rel_hum_stats = StatsContainer("rel_hum_sensor")
    for packet in wrapped_packets:
        rel_hum = packet.get_sensors().get_relative_humidity()
        if rel_hum and packet.get_sensors().validate_relative_humidity():
            data_list.extend(rel_hum.get_samples().get_values())
            ts = rel_hum.get_timestamps().get_timestamps()
            timestamps.extend(ts)
            if rel_hum.get_timestamps().get_timestamps_count() == 1:
                rel_hum_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                rel_hum_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list) > 0:
        return load_apim_single_sensor(SensorType.RELATIVE_HUMIDITY, timestamps, data_list, gaps, "rel_humidity",
                                       get_sensor_description_list(wrapped_packets, SensorType.RELATIVE_HUMIDITY),
                                       rel_hum_stats.mean_of_means())
    return None


def load_apim_accelerometer(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load accelerometer data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: accelerometer sensor data if it exists, None otherwise
    """
    accel = wrapped_packet.get_sensors().get_accelerometer()
    if accel and wrapped_packet.get_sensors().validate_accelerometer():
        data_df = read_apim_xyz_sensor(accel, "accelerometer")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(accel),
            data_df,
            SensorType.ACCELEROMETER,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_accelerometer_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                      gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load accelerometer data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: accelerometer sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    accel_stats = StatsContainer("accel_sensor")
    for packet in wrapped_packets:
        accel = packet.get_sensors().get_accelerometer()
        if accel and packet.get_sensors().validate_accelerometer():
            ts = accel.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(accel.get_x_samples().get_values())
            data_list[2].extend(accel.get_y_samples().get_values())
            data_list[3].extend(accel.get_z_samples().get_values())
            if accel.get_timestamps().get_timestamps_count() == 1:
                accel_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                accel_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.ACCELEROMETER, data_list, gaps, "accelerometer",
                                    get_sensor_description_list(wrapped_packets, SensorType.ACCELEROMETER),
                                    accel_stats.mean_of_means())
    return None


def load_apim_magnetometer(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load magnetometer data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: magnetometer sensor data if it exists, None otherwise
    """
    mag = wrapped_packet.get_sensors().get_magnetometer()
    if mag and wrapped_packet.get_sensors().validate_magnetometer():
        data_df = read_apim_xyz_sensor(mag, "magnetometer")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(mag),
            data_df,
            SensorType.MAGNETOMETER,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_magnetometer_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                     gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load magnetometer data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: magnetometer sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    mag_stats = StatsContainer("mag_sensor")
    for packet in wrapped_packets:
        mag = packet.get_sensors().get_magnetometer()
        if mag and packet.get_sensors().validate_magnetometer():
            ts = mag.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(mag.get_x_samples().get_values())
            data_list[2].extend(mag.get_y_samples().get_values())
            data_list[3].extend(mag.get_z_samples().get_values())
            if mag.get_timestamps().get_timestamps_count() == 1:
                mag_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                mag_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.MAGNETOMETER, data_list, gaps, "magnetometer",
                                    get_sensor_description_list(wrapped_packets, SensorType.MAGNETOMETER),
                                    mag_stats.mean_of_means())
    return None


def load_apim_gyroscope(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load gyroscope data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: gyroscope sensor data if it exists, None otherwise
    """
    gyro = wrapped_packet.get_sensors().get_gyroscope()
    if gyro and wrapped_packet.get_sensors().validate_gyroscope():
        data_df = read_apim_xyz_sensor(gyro, "gyroscope")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(gyro),
            data_df,
            SensorType.GYROSCOPE,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_gyroscope_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                  gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load gyroscope data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: gyroscope sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    gyro_stats = StatsContainer("gyro_sensor")
    for packet in wrapped_packets:
        gyro = packet.get_sensors().get_gyroscope()
        if gyro and packet.get_sensors().validate_gyroscope():
            ts = gyro.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(gyro.get_x_samples().get_values())
            data_list[2].extend(gyro.get_y_samples().get_values())
            data_list[3].extend(gyro.get_z_samples().get_values())
            if gyro.get_timestamps().get_timestamps_count() == 1:
                gyro_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                gyro_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.GYROSCOPE, data_list, gaps, "gyroscope",
                                    get_sensor_description_list(wrapped_packets, SensorType.GYROSCOPE),
                                    gyro_stats.mean_of_means())
    return None


def load_apim_gravity(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load gravity data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: gravity sensor data if it exists, None otherwise
    """
    gravity = wrapped_packet.get_sensors().get_gravity()
    if gravity and wrapped_packet.get_sensors().validate_gravity():
        data_df = read_apim_xyz_sensor(gravity, "gravity")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(gravity),
            data_df,
            SensorType.GRAVITY,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_gravity_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load gravity data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: gravity sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    gravity_stats = StatsContainer("gravity_sensor")
    for packet in wrapped_packets:
        gravity = packet.get_sensors().get_gravity()
        if gravity and packet.get_sensors().validate_gravity():
            ts = gravity.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(gravity.get_x_samples().get_values())
            data_list[2].extend(gravity.get_y_samples().get_values())
            data_list[3].extend(gravity.get_z_samples().get_values())
            if gravity.get_timestamps().get_timestamps_count() == 1:
                gravity_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                gravity_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.GRAVITY, data_list, gaps, "gravity",
                                    get_sensor_description_list(wrapped_packets, SensorType.GRAVITY),
                                    gravity_stats.mean_of_means())
    return None


def load_apim_orientation(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load orientation data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: orientation sensor data if it exists, None otherwise
    """
    orientation = wrapped_packet.get_sensors().get_orientation()
    if orientation and wrapped_packet.get_sensors().validate_orientation():
        data_df = read_apim_xyz_sensor(orientation, "orientation")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(orientation),
            data_df,
            SensorType.ORIENTATION,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_orientation_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                    gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load orientation data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: orientation sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    orient_stats = StatsContainer("orient_sensor")
    for packet in wrapped_packets:
        orient = packet.get_sensors().get_orientation()
        if orient and packet.get_sensors().validate_orientation():
            ts = orient.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(orient.get_x_samples().get_values())
            data_list[2].extend(orient.get_y_samples().get_values())
            data_list[3].extend(orient.get_z_samples().get_values())
            if orient.get_timestamps().get_timestamps_count() == 1:
                orient_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                orient_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.ORIENTATION, data_list, gaps, "orientation",
                                    get_sensor_description_list(wrapped_packets, SensorType.ORIENTATION),
                                    orient_stats.mean_of_means())
    return None


def load_apim_linear_accel(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load linear acceleration data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: linear acceleration sensor data if it exists, None otherwise
    """
    linear_accel = wrapped_packet.get_sensors().get_linear_acceleration()
    if linear_accel and wrapped_packet.get_sensors().validate_linear_acceleration():
        data_df = read_apim_xyz_sensor(linear_accel, "linear_accel")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(linear_accel),
            data_df,
            SensorType.LINEAR_ACCELERATION,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_linear_accel_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                     gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load linear acceleration data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: linear acceleration sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    lin_acc_stats = StatsContainer("lin_acc_sensor")
    for packet in wrapped_packets:
        lin_acc = packet.get_sensors().get_linear_acceleration()
        if lin_acc and packet.get_sensors().validate_linear_acceleration():
            ts = lin_acc.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(lin_acc.get_x_samples().get_values())
            data_list[2].extend(lin_acc.get_y_samples().get_values())
            data_list[3].extend(lin_acc.get_z_samples().get_values())
            if lin_acc.get_timestamps().get_timestamps_count() == 1:
                lin_acc_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                lin_acc_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.LINEAR_ACCELERATION, data_list, gaps, "linear_accel",
                                    get_sensor_description_list(wrapped_packets, SensorType.LINEAR_ACCELERATION),
                                    lin_acc_stats.mean_of_means())
    return None


def load_apim_rotation_vector(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load rotation vector data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: rotation vector sensor data if it exists, None otherwise
    """
    rotation = wrapped_packet.get_sensors().get_rotation_vector()
    if rotation and wrapped_packet.get_sensors().validate_rotation_vector():
        data_df = read_apim_xyz_sensor(rotation, "rotation_vector")
        sample_rate, sample_interval, sample_interval_std = get_sample_statistics(data_df)
        return SensorData(
            get_sensor_description(rotation),
            data_df,
            SensorType.ROTATION_VECTOR,
            sample_rate,
            sample_interval,
            sample_interval_std,
            False,
        )


def load_apim_rotation_vector_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                                        gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load rotation vector data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: rotation vector sensor data if it exists, None otherwise
    """
    data_list = [[], [], [], []]
    rot_vec_stats = StatsContainer("rot_vec_sensor")
    for packet in wrapped_packets:
        rot_vec = packet.get_sensors().get_rotation_vector()
        if rot_vec and packet.get_sensors().validate_rotation_vector():
            ts = rot_vec.get_timestamps().get_timestamps()
            data_list[0].extend(ts)
            data_list[1].extend(rot_vec.get_x_samples().get_values())
            data_list[2].extend(rot_vec.get_y_samples().get_values())
            data_list[3].extend(rot_vec.get_z_samples().get_values())
            if rot_vec.get_timestamps().get_timestamps_count() == 1:
                rot_vec_stats.add(dtu.seconds_to_microseconds(packet.get_packet_duration_s()), 0, 1)
            else:
                rot_vec_stats.add(np.mean(np.diff(ts)), np.std(np.diff(ts)), len(ts) - 1)
    if len(data_list[0]) > 0:
        return load_apim_xyz_sensor(SensorType.ROTATION_VECTOR, data_list, gaps, "rotation_vector",
                                    get_sensor_description_list(wrapped_packets, SensorType.ROTATION_VECTOR),
                                    rot_vec_stats.mean_of_means())
    return None


def load_apim_health(wrapped_packet: WrappedRedvoxPacketM) -> Optional[SensorData]:
    """
    load station health data from a single wrapped packet
    :param wrapped_packet: packet with data to load
    :return: station health data if it exists, None otherwise
    """
    metrics = wrapped_packet.get_station_information().get_station_metrics()
    timestamps = metrics.get_timestamps().get_timestamps()
    if len(timestamps) > 0:
        bat_samples = metrics.get_battery().get_values()
        bat_cur_samples = metrics.get_battery_current().get_values()
        temp_samples = metrics.get_temperature().get_values()
        net_samples = metrics.get_network_type().get_values()
        net_str_samples = metrics.get_network_strength().get_values()
        pow_samples = metrics.get_power_state().get_values()
        avail_ram_samples = metrics.get_available_ram().get_values()
        avail_disk_samples = metrics.get_available_disk().get_values()
        cell_samples = metrics.get_cell_service_state().get_values()
        data_for_df = []
        for i in range(len(timestamps)):
            new_entry = [
                timestamps[i],
                timestamps[i],
                np.nan if len(bat_samples) < i + 1 else bat_samples[i],
                np.nan if len(bat_cur_samples) < i + 1 else bat_cur_samples[i],
                np.nan if len(temp_samples) < i + 1 else temp_samples[i],
                np.nan if len(net_samples) < i + 1 else net_samples[i],
                np.nan if len(net_str_samples) < i + 1 else net_str_samples[i],
                np.nan if len(pow_samples) < i + 1 else pow_samples[i],
                np.nan if len(avail_ram_samples) < i + 1 else avail_ram_samples[i],
                np.nan if len(avail_disk_samples) < i + 1 else avail_disk_samples[i],
                np.nan if len(cell_samples) < i + 1 else cell_samples[i],
            ]
            data_for_df.append(new_entry)
        data_df = pd.DataFrame(
            data_for_df,
            columns=[
                "timestamps",
                "unaltered_timestamps",
                "battery_charge_remaining",
                "battery_current_strength",
                "internal_temp_c",
                "network_type",
                "network_strength",
                "power_state",
                "avail_ram",
                "avail_disk",
                "cell_service",
            ],
        )
        if len(timestamps) > 1:
            sample_rate = 1
            sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(data_df["timestamps"]))))
        else:
            sample_rate = 1 / wrapped_packet.get_packet_duration_s()
            sample_interval_std = np.nan
        return SensorData(
            "station health",
            data_df,
            SensorType.STATION_HEALTH,
            sample_rate,
            1 / sample_rate,
            sample_interval_std
            )
    return None


def load_apim_health_from_list(wrapped_packets: List[WrappedRedvoxPacketM],
                               gaps: List[Tuple[float, float]]) -> Optional[SensorData]:
    """
    load station health data from a list of wrapped packets
    :param wrapped_packets: packets with data to load
    :param gaps: the list of non-inclusive start and end times of the gaps in the packets
    :return: station health sensor data if it exists, None otherwise
    """
    data_list: List = [[], [], [], [], [], [], [], [], [], []]
    for packet in wrapped_packets:
        metrics = packet.get_station_information().get_station_metrics()
        timestamps = metrics.get_timestamps().get_timestamps()
        num_samples = len(timestamps)
        if num_samples > 0:
            data_list[0].extend(timestamps)
            samples = metrics.get_battery().get_values()
            if len(samples) != num_samples:
                samples = np.full(num_samples, np.nan)
            data_list[1].extend(samples)
            samples = metrics.get_battery_current().get_values()
            if len(samples) != num_samples:
                samples = np.full(num_samples, np.nan)
            data_list[2].extend(samples)
            samples = metrics.get_temperature().get_values()
            if len(samples) != num_samples:
                samples = np.full(num_samples, np.nan)
            data_list[3].extend(samples)
            samples = metrics.get_network_type().get_values()
            data_list[4].extend([NetworkType["UNKNOWN_NETWORK"]
                                 if len(samples) < i + 1 else samples[i]
                                 for i in range(num_samples)])
            samples = metrics.get_network_strength().get_values()
            if len(samples) != num_samples:
                samples = np.full(num_samples, np.nan)
            data_list[5].extend(samples)
            samples = metrics.get_power_state().get_values()
            data_list[6].extend([PowerState["UNKNOWN_POWER_STATE"]
                                 if len(samples) < i + 1 else samples[i]
                                 for i in range(num_samples)])
            samples = metrics.get_available_ram().get_values()
            if len(samples) != num_samples:
                samples = np.full(num_samples, np.nan)
            data_list[7].extend(samples)
            samples = metrics.get_available_disk().get_values()
            data_list[8].extend(samples if len(samples) == num_samples else np.full(num_samples, np.nan))
            samples = metrics.get_cell_service_state().get_values()
            data_list[9].extend([CellServiceState["UNKNOWN"]
                                 if len(samples) < i + 1 else samples[i]
                                 for i in range(num_samples)])
    if len(data_list[0]) > 0:
        data_list.insert(1, data_list[0].copy())
        # health is collected 1 per packet or 1 per second
        if len(data_list[0]) > len(wrapped_packets):
            sample_rate = 1.0
        else:
            sample_rate = 1 / wrapped_packets[0].get_packet_duration_s()
        sample_interval = 1 / sample_rate
        sample_interval_std = dtu.microseconds_to_seconds(float(np.std(np.diff(data_list[0])))) \
            if len(data_list[0]) > 1 else np.nan
        df = gpu.fill_gaps(pd.DataFrame(
            np.transpose(data_list),
            columns=["timestamps", "unaltered_timestamps", "battery_charge_remaining", "battery_current_strength",
                     "internal_temp_c", "network_type", "network_strength", "power_state", "avail_ram", "avail_disk",
                     "cell_service"],
        ), gaps, dtu.seconds_to_microseconds(sample_interval))
        return SensorData(
            "station health",
            df,
            SensorType.STATION_HEALTH,
            sample_rate,
            sample_interval,
            sample_interval_std
        )
    return None



"""
This module contains functions for downloading data from AWS S3 via signed URLs.
"""

import logging
import os
from typing import Tuple, Optional

import requests

# pylint: disable=C0103
log = logging.getLogger(__name__)


def find_between(start: str, end: str, contents: str) -> str:
    """
    Find the string contents between two other strings.
    :param start: The first string.
    :param end: The seconds string.
    :param contents: The full string.
    :return: The contents between the start and end strings.
    """
    s_idx = contents.find(start)
    e_idx = contents.find(end)
    return contents[s_idx + len(start) : e_idx]


def get_file(
    url: str, retries: int, session: requests.Session = requests.Session()
) -> Optional[bytes]:
    """
    Attempts to download a file with a configurable amount of retries.
    :param url: The url to download.
    :param retries: Number of retries.
    :param session: An instance of a session.
    :return: The bytes of the file.
    """
    # pylint: disable=W0702
    # noinspection PyBroadException
    try:
        resp: requests.Response = session.get(url)
        if resp.status_code == 200:
            return resp.content
        else:
            log.error(
                "Received error response when requesting data for url=%s: %d %s",
                url,
                resp.status_code,
                resp.text,
            )
            if retries > 0:
                log.info("Retrying with %d retries", retries)
                return get_file(url, retries - 1, session)

            log.error("All retries exhausted, could not get %s", url)
            return None
    except Exception as e:
        log.error("Encountered an error while getting data for %s: %s", url, str(e))
        if retries > 0:
            log.info("Retrying with %d retries", retries)
            return get_file(url, retries - 1, session)
        log.error("All retries exhausted, could not get %s", url)
        return None


def download_file(
    url: str, session: requests.Session, out_dir: str, retries: int
) -> Tuple[str, int]:
    """
    Attempts to download a file from S3.
    :param url: The URL to retrieve.
    :param session: The HTTP session.
    :param out_dir: The output directory where files will be stored.
    :param retries: The number of times to retry failed file downloads.
    :return: A tuple containing the data_key and the length of the download response.
    """
    buf: Optional[bytes] = get_file(url, retries, session)

    if buf:
        data_key = find_between("/rdvxdata/", "?X-Amz-Algorithm=", url)

        directory = os.path.dirname(data_key)
        full_dir = f"{out_dir}/{directory}"
        if not os.path.exists(full_dir):
            log.debug("Directory %s does not exist, creating it", full_dir)
            os.makedirs(full_dir)

        full_path = f"{out_dir}/{data_key}"
        with open(full_path, "wb") as fout:
            bytes_written = fout.write(buf)
            log.debug("Wrote %s %d", full_path, bytes_written)

        return data_key, len(buf)

    return "", 0
